﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class FABRIK : MonoBehaviour
{
    public Transform target;

    List<Transform> joints;
    Vector3[] positions;
    float[] distances;
    float totalDist;

    // Collect all joints recursively, order in "joints" keeps the order in hierarchy
    void FindJoints(Transform t, ref List<Transform> joints)
    {
        foreach (Transform child in t)
        {
            if (child.name.Contains("Joint"))
            {
                joints.Add(child);
            }
            if (child.name.Contains("Joint") || child.name.Contains("Bone"))
            {
                FindJoints(child.gameObject.transform, ref joints);
            }
        }
    }

    void Start()
    {
        joints = new List<Transform>();
        FindJoints(transform, ref joints);
        if (joints.Count == 0)
        {
            Debug.LogWarning("No joints found");
            enabled = false;
            return;
        }
        positions = new Vector3[joints.Count];
        distances = new float[joints.Count - 1];
        int i = 0;
        foreach(Transform joint in joints)
        {
            positions[i] = joint.position;
            ++i;
            //Debug.Log(joint.name + " -> " + joint.position);
        }
        totalDist = 0;
        for (int j = 0; j < distances.Length; ++j)
        {
            distances[j] = Vector3.Distance(positions[j + 1], positions[j]);
            totalDist += distances[j];
        }
    }

    bool Reachable()
    {
        return totalDist > Vector3.Distance(transform.position, target.transform.position);
    }

    void UpdateJoints()
    {
        int n = positions.Length;
        for (int i = 0; i < n-1; ++i)
        {
            Quaternion parentRotation = Quaternion.identity;
            if (i > 0) // has parent
            {
                parentRotation = joints[i - 1].rotation;
            }
            Vector3 baseDir = Vector3.up;
            Vector3 newDir = Quaternion.Inverse(parentRotation) * (positions[i + 1] - positions[i]);
            Vector3 rotAxis = Vector3.Cross(baseDir.normalized, newDir.normalized).normalized;
            float rotAngle = Vector3.SignedAngle(baseDir.normalized, newDir.normalized, rotAxis);
            if (rotAngle < 0.001)
            {
                joints[i].transform.localRotation = Quaternion.identity;
            }
            else
            {
                joints[i].transform.localRotation = Quaternion.AngleAxis(rotAngle, rotAxis);
            }
        }
    }

    void Update()
    {
        if (target == null) return;

        int n = positions.Length;
        Vector3 t = target.transform.position;
        if (Reachable())
        {
            Vector3 b = positions[0];
            float dif = Vector3.Distance(positions[n - 1], t);
            const float difThreshold = 0.01f;
            int iter = 0;
            while (dif > difThreshold && iter++ < 100)
            {
                positions[n - 1] = t;
                for (int i = n-2; i >= 0; --i)
                {
                    float r = Vector3.Distance(positions[i + 1], positions[i]);
                    float lambda = distances[i] / r;
                    positions[i] = (1.0f - lambda) * positions[i + 1] + lambda * positions[i];
                }

                positions[0] = b;
                for (int i = 0; i < n-1; ++i)
                {
                    float r = Vector3.Distance(positions[i + 1], positions[i]);
                    float lambda = distances[i] / r;
                    positions[i + 1] = (1.0f - lambda) * positions[i] + lambda * positions[i + 1];
                }
            }
        }
        else
        {
            for (int i = 0; i < n-1; ++i)
            {
                float r = Vector3.Distance(t, positions[i]);
                float lambda = distances[i] / r;
                positions[i + 1] = (1.0f - lambda) * positions[i] + lambda * t;
            }
        }
        UpdateJoints();
    }
}
